#ifndef CKCoreQRCodeGenerator_h
#define CKCoreQRCodeGenerator_h

#import <Foundation/Foundation.h>

#if TARGET_OS_OSX
  #include <CodKits_macOS/CodKits_macOS.h>
#else
  #include <CodKits_iOS/CodKits_iOS.h>
#endif

/// The levels of error correction available for QR code generation.
typedef NS_ENUM(NSUInteger, QRErrorCorrection)
  {
  /// Low error correction level (7%).
  QRErrorCorrectionLow      = 0,
  /// Medium error correction level (15%).
  QRErrorCorrectionMedium   = 1,
  /// Quartile error correction level (25%).
  QRErrorCorrectionQuartile = 2,
  /// High error correction level (30%).
  QRErrorCorrectionHigh     = 3,
  };

//-------------------------------------------------------------------------------------------------------------------------
// INTERFACE
//-------------------------------------------------------------------------------------------------------------------------

/// A class that generates QR codes with customizable properties such as size, color, and error correction level.
///
/// Example usage:
/// @code
/// CKCoreQRCodeGenerator *qr = [[CKCoreQRCodeGenerator alloc] initWithString:@"Hello World"];
/// qr.size = CGSizeMake(400.0f, 400.0f); // 400x400 size
/// qr.color = [CIColor colorWithRGBA:@"#FFFFFF"]; // white QR Code color
/// qr.backgroundColor = [CIColor colorWithRGBA:@"#000000"]; // black background color
/// NSImage *qrImage = [qr getImage];
/// imageView.image = qrImage;
/// @endcode
@interface CKCoreQRCodeGenerator : NSObject

/// The data used to generate the QR code.
@property (nonatomic, strong, readonly) NSData *data;

/// The error correction level of the QR code. Default is `QRErrorCorrectionMedium`.
@property (nonatomic, assign) QRErrorCorrection errorCorrection;

/// The size of the QR code to be generated.
@property (nonatomic, assign) CGSize size;

/// The color of the QR code. Defaults to black.
@property (nonatomic, strong) CIColor *color;

/// The background color of the QR code. Defaults to white.
@property (nonatomic, strong) CIColor *backgroundColor;

/// Initializes a new `CKCoreQRCodeGenerator` object with default properties.
/// @return A new instance of `CKCoreQRCodeGenerator`.
- (instancetype)init;

/// Initializes a new `CKCoreQRCodeGenerator` object with the provided data.
/// @param data The data to be encoded into the QR code.
/// @return A new instance of `CKCoreQRCodeGenerator`.
- (instancetype)initWithData:(NSData *)data;

/// Initializes a new `CKCoreQRCodeGenerator` object with the provided string.
/// @param string The string to be encoded into the QR code.
/// @return A new instance of `CKCoreQRCodeGenerator`.
- (instancetype)initWithString:(NSString *)string;

/// Generates and returns the QR code as an `NSImage`.
/// @return An `NSImage` representing the generated QR code.
- (NSImage *)getImage;

@end

#endif /* CKCoreQRCodeGenerator_h */

